//LED Matrix Controller w/ Arduino
// Adam Greig 2008
// CC BY-SA-NC 3.0

#define F_CPU 12000000L
#define CPU_FREQ 12000000L

#include <WProgram.h>
#include <avr/interrupt.h>
#include <avr/io.h>

#include "font.h"

volatile short int matrix_reds[8];
volatile short int matrix_greens[8];

//Write out an array to the 3 shift registers
void writeMatrix() {
	int i,j;
	
	//For each cathode...
	for( i=0; i<8; i++ ) {
		
		//Greens:
		//Latch pin low (registers start taking data)
		PORTB	= B000;
		
		//Output cathodes
		for( j=0; j<8; j++ ) {
			PORTB	= B000;
			//I suck and wired the output for cathode 5 to cathode 8 on the matrix, 6 to 7, 7 to 6, 8 to 5.
			//  Since it's a PCB, I can't fix it in hardware...
			//  That's what software is for!
			if( i < 4 ) {
				if( j == (7 - i) )
					PORTB	= B010;
			} else {
				if( j == (i - 4) )
					PORTB	= B010;
			}
			PORTB	= PORTB | B100;
		}
		
		//Output empty reds
		for( j=0; j<8; j++ ) {
			PORTB	= B000;
			PORTB	= B100;
		}
		
		//Output greens
		for( j=8; j<16; j++ ) {
			PORTB	= B000;
			if( matrix_greens[7 - i] & ( B1 << j ) )
				PORTB	= B010;
			PORTB	= PORTB | B100;
		}
		
		//Output two empty rows of anodes for now
		for( j=0; j<16; j++ ) {
			PORTB	= B000;
			PORTB	= B100;
		}
		
		//Latch pin high (registers output data)
		PORTB	= B001;
		
		delayMicroseconds( 150 );
		
		//Reds:
		//Latch pin low (registers start taking data)
		PORTB	= B000;
		
		//Output cathodes
		for( j=0; j<8; j++ ) {
			PORTB	= B000;
			//I suck and wired the output for cathode 5 to cathode 8 on the matrix, 6 to 7, 7 to 6, 8 to 5.
			//  Since it's a PCB, I can't fix it in hardware...
			//  That's what software is for!
			if( i < 4 ) {
				if( j == (7 - i) )
					PORTB	= B010;
			} else {
				if( j == (i - 4) )
					PORTB	= B010;
			}
			PORTB	= PORTB | B100;
		}
		
		//Output reds
		for( j=8; j<16; j++ ) {
			PORTB	= B000;
			if( matrix_reds[7 - i] & ( B1 << j ) )
				PORTB	= B010;
			PORTB	= PORTB | B100;
		}
		
		//Output empty greens
		for( j=0; j<8; j++ ) {
			PORTB	= B000;
			PORTB	= B100;
		}
		
		//Output two empty rows of anodes for now
		for( j=0; j<16; j++ ) {
			PORTB	= B000;
			PORTB	= B100;
		}
		
		//Latch pin high (registers output data)
		PORTB	= B001;
		
		delayMicroseconds( 150 );
		
		PORTB	= B000;
		//Output a complete blank
		for( j=0; j<40; j++ ) {
			PORTB	= B000;
			PORTB	= B100;
		}
		PORTB	= B001;
	}
	
}

void setup() {
	Serial.begin( 300 );
	
	//All 3 wires to outputs
	DDRB	= B111;
	
	//Timer2 Settings: Timer Prescaler /256, WGM mode 0
	TCCR2A = 0;
	TCCR2B = 1<<CS22 | 1<<CS21; 
	
	//Timer2 Overflow Interrupt Enable
	TIMSK2 = 1<<TOIE2;
	
	//reset timer
	TCNT2 = 0;
	
	for( int i = 0; i < 8; i++ ) {
		matrix_reds[i] = 0;
		matrix_greens[i] = 0;
	}
	
}


void loop() {
	
	char msg[80] = "";
	int msg_index = 0;
	int msg_letter = 0;
	int msg_counter;
	
	int ser_data;
	
	volatile short int* matrix = matrix_reds;
	
	for( int i = 0;; i++ ) {
		
		if( Serial.available() ) {
			if( Serial.read() == 0x02 ) {
				msg_counter = 0;
				msg_index = 0;
				ser_data = 0;
				
				while( ser_data != 0x03 ) {
					if( Serial.available() ) {
						ser_data = Serial.read();
						msg[msg_counter++] = ser_data;
					}
				}
				
				msg[msg_counter - 1] = 0x00;
				msg_letter = msg[0];
				
			}
		}
		
		for( int j = 0; j < 8; j++ ) {
			matrix_reds[j] <<= 1;
			matrix_greens[j] <<= 1;
		}
		
		if( i == 6 ) {
			msg_letter = msg[ msg_index ];
			if( msg_letter == 0 ) {
				msg_index = 0;
				msg_letter = msg[ 0 ];
				if( matrix == matrix_reds )
					matrix = matrix_greens;
				else
					matrix = matrix_reds;
			}
			msg_index++;
			if( msg_letter == 0 ) msg_letter = 0x20;
			for( int j = 0; j < 8; j++ ) {
				matrix[j] |= font[msg_letter - 32][j];
			}
			i = 0;
		}
		delay( 50 );
	}
}


//Timer2 overflow interrupt vector handler, called (12,000,000/256)/256 times per second
ISR(TIMER2_OVF_vect) {
	writeMatrix();
};

int main() {
	init();
	setup();
	for(;;)
		loop();
	return 0;
}
